package org.jboss.cache;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.buddyreplication.BuddyFqnTransformer;
import org.jboss.cache.buddyreplication.BuddyManager;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.marshall.NodeData;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.transaction.GlobalTransaction;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * A container for the root node in the cache, which also provides helpers for efficiently accessing nodes, walking trees, etc.
 *
 * @author Mircea.Markus@jboss.com
 * @since 2.2
 */
public class DataContainer
{
   private static final Log log = LogFactory.getLog(DataContainer.class);
   private static boolean trace = log.isTraceEnabled();

   private Configuration configuration;

   /**
    * Root node.
    */
   private NodeSPI root;

   /**
    * Set<Fqn> of Fqns of the topmost node of internal regions that should
    * not included in standard state transfers.
    */
   private final Set<Fqn> internalFqns = new HashSet<Fqn>();

   @Inject
   public void injectDependencies(Configuration configuration)
   {
      this.configuration = configuration;
   }

   /**
    * Retrieves the root node.
    *
    * @return the root node
    */
   public NodeSPI getRoot()
   {
      return root;
   }

   /**
    * Sets the root node reference to the node passed in.
    *
    * @param root node
    */
   public void setRoot(NodeSPI root)
   {
      if (root == null || !root.getFqn().isRoot())
         throw new CacheException("Attempting to set an invalid node [" + root + "] as a root node!");
      this.root = root;
   }

   /**
    * Adds the specified Fqn to the list of Fqns to be considered "internal".
    *
    * @param fqn fqn to add to list
    */
   public void registerInternalFqn(Fqn fqn)
   {
      internalFqns.add(fqn);
   }

   /**
    * Finds a node given a fully qualified name, directly off the interceptor chain.  In the event of an exception,
    * returns null.  Does not include invalid or deleted nodes.
    *
    * @param fqn Fully qualified name for the corresponding node.
    * @return Node referenced by the given Fqn, or null if the node cannot be found or if there is an exception.
    */
   public NodeSPI peek(Fqn fqn)
   {
      try
      {
         return peek(fqn, null);
      }
      catch (CacheException e)
      {
         log.warn("Unexpected error", e);
         return null;
      }
   }

   /**
    * Finds a node given a fully qualified name and DataVersion.  Does not include invalid or deleted nodes.  If the data
    * version passed in is null, then data version checking is skipped.  Data version checking is also skipped if optimistic
    * locking is not used.
    *
    * @param fqn     fqn to find
    * @param version version of the node to find
    * @return a node, if found, or null if not.
    */
   public NodeSPI peek(Fqn fqn, DataVersion version)
   {
      return peek(fqn, version, false);
   }

   /**
    * Similar to {@link #peek(Fqn, org.jboss.cache.optimistic.DataVersion)} except that it throws a {@link org.jboss.cache.NodeNotExistsException}
    * if the node cannot be found.
    *
    * @param gtx            global transaction
    * @param fqn            fqn to find
    * @param includeInvalid if true, invalid nodes are considered as well.
    * @return the node
    */
   public NodeSPI peekStrict(GlobalTransaction gtx, Fqn fqn, boolean includeInvalid)
   {
      NodeSPI n = peek(fqn, null, includeInvalid);
      if (n == null)
      {
         StringBuilder builder = new StringBuilder();
         builder.append("Node ").append(fqn).append(" not found (gtx=").append(gtx).append(")");
         String errStr = builder.toString();
         if (trace) log.trace(errStr);
         throw new NodeNotExistsException(errStr);
      }
      return n;
   }

   /**
    * Searches for a specific node, with a specific data version and, optionally, invalid nodes as well, but not
    * deleted nodes.  If the data version passed in is null, then data version checking is skipped.  Data version
    * checking is also skipped if optimistic locking is not used.
    *
    * @param fqn                 Fqn to find
    * @param version             version of the node to find
    * @param includeInvalidNodes if true, invalid nodes are considered
    * @return the node, if found, or null otherwise.
    */
   public NodeSPI peek(Fqn fqn, DataVersion version, boolean includeInvalidNodes)
   {
      if (fqn == null) return null;

      NodeSPI toReturn = peek(fqn, false, includeInvalidNodes);

      if (toReturn != null && version != null && configuration.isNodeLockingOptimistic())
      {
         // we need to check the version of the data node...
         DataVersion nodeVersion = toReturn.getVersion();
         if (trace)
         {
            log.trace("looking for optimistic node [" + fqn + "] with version [" + version + "].  My version is [" + nodeVersion + "]");
         }
         if (nodeVersion.newerThan(version))
         {
            // we have a versioning problem; throw an exception!
            throw new CacheException("Unable to validate versions.");
         }
      }
      return toReturn;
   }

   /**
    * Same as calling <tt>peek(fqn, includeDeletedNodes, false)</tt>.
    *
    * @param fqn                 Fqn to find
    * @param includeDeletedNodes if true, deleted nodes are considered
    * @return the node, if found, or null otherwise.
    */
   public NodeSPI peek(Fqn<?> fqn, boolean includeDeletedNodes)
   {
      return peek(fqn, includeDeletedNodes, false);
   }

   /**
    * Peeks for a specified node.  This involves a direct walk of the tree, starting at the root, until the required node
    * is found.  If the node is not found, a null is returned.
    *
    * @param fqn                 Fqn of the node to find
    * @param includeDeletedNodes if true, deleted nodes are also considered
    * @param includeInvalidNodes if true, invalid nodes are also considered
    * @return the node, if found, or null otherwise.
    */
   public NodeSPI peek(Fqn<?> fqn, boolean includeDeletedNodes, boolean includeInvalidNodes)
   {
      if (fqn == null || fqn.size() == 0) return getRoot();
      NodeSPI n = getRoot();
      int fqnSize = fqn.size();
      for (int i = 0; i < fqnSize; i++)
      {
         Object obj = fqn.get(i);
         n = n.getChildDirect(obj);
         if (n == null)
         {
            return null;
         }
         else if (!includeDeletedNodes && n.isDeleted())
         {
            return null;
         }
         else if (!includeInvalidNodes && !n.isValid())
         {
            return null;
         }
      }
      return n;
   }

   /**
    * Tests if an Fqn exists and is valid and not deleted.
    *
    * @param fqn the fqn representing the node to test
    * @return true if the node exists, false otherwise.
    */
   public boolean exists(Fqn fqn)
   {
      return peek(fqn, false, false) != null;
   }

   /**
    * Returns true if the Fqn exists, is valid and is not deleted, and the node has children.
    *
    * @param fqn the fqn to test
    * @return true if the Fqn exists, is valid and is not deleted, and the node has children.
    */
   public boolean hasChildren(Fqn fqn)
   {
      if (fqn == null) return false;

      NodeSPI n = peek(fqn);
      return n != null && n.hasChildrenDirect();
   }

   /**
    * Prepares a list of {@link NodeData} objects for a specified node and all its children.
    *
    * @param list List of NodeData objects, which will be added to.
    * @param node node to recursively add to the list
    * @return the same list passed in
    */
   public List<NodeData> buildNodeData(List<NodeData> list, NodeSPI node)
   {
      NodeData data = new NodeData(BuddyFqnTransformer.getActualFqn(node.getFqn()), node.getDataDirect());
      list.add(data);
      for (Object childNode : node.getChildrenDirect())
      {
         buildNodeData(list, (NodeSPI) childNode);
      }
      return list;
   }

   /**
    * Generates a list of nodes for eviction.  This filters out nodes that cannot be evicted, such as those which are
    * marked as resident.  See {@link NodeSPI#setResident(boolean)}.
    *
    * @param fqn       the node to consider for eviction
    * @param recursive if recursive, child nodes are also considered
    * @return a list of Fqns that can be considered for eviction
    */
   public List<Fqn> getNodesForEviction(Fqn fqn, boolean recursive)
   {
      List<Fqn> result = new ArrayList<Fqn>();
      NodeSPI node = peek(fqn, false);
      if (recursive)
      {
         recursiveAddEvictionNodes(node, result);
      }
      else
      {
         if (node == null)
         {
            result.add(fqn);
            return result;
         }
         buildNodesForEviction(node, result);
      }
      return result;
   }

   private void recursiveAddEvictionNodes(NodeSPI node, List<Fqn> result)
   {
      for (NodeSPI child : (Set<NodeSPI>) node.getChildrenDirect())
      {
         recursiveAddEvictionNodes(child, result);
      }
      buildNodesForEviction(node, result);
   }

   private void buildNodesForEviction(NodeSPI node, List<Fqn> nodes)
   {
      if (node == null || node.isResident())
      {
         return;
      }
      Fqn fqn = node.getFqn();
      if (fqn.isRoot())
      {
         for (Object childName : node.getChildrenNamesDirect())
         {
            if (!node.isResident()) nodes.add(Fqn.fromRelativeElements(fqn, childName));
         }
      }
      else
      {
         nodes.add(fqn);
      }
   }

   @Override
   public String toString()
   {
      return toString(false);
   }

   /**
    * Returns a Set<Fqn> of Fqns of the topmost node of internal regions that
    * should not included in standard state transfers. Will include
    * {@link BuddyManager#BUDDY_BACKUP_SUBTREE} if buddy replication is
    * enabled.
    *
    * @return an unmodifiable Set<Fqn>.  Will not return <code>null</code>.
    */
   public Set<Fqn> getInternalFqns()
   {
      return Collections.unmodifiableSet(internalFqns);
   }

   /**
    * Returns a debug string with optional details of contents.
    */
   public String toString(boolean details)
   {
      StringBuffer sb = new StringBuffer();
      int indent = 0;

      if (!details)
      {
         sb.append(getClass().getName()).append(" [").append(getNumberOfNodes()).append(" nodes, ");
         sb.append(getNumberOfLocksHeld()).append(" locks]");
      }
      else
      {
         if (root == null)
            return sb.toString();
         for (Object n : root.getChildrenDirect())
         {
            ((NodeSPI) n).print(sb, indent);
            sb.append("\n");
         }
      }
      return sb.toString();
   }

   /**
    * Returns the number of read or write locks held across the entire cache.
    */
   public int getNumberOfLocksHeld()
   {
      return numLocks(root);
   }

   private int numLocks(NodeSPI n)
   {
      int num = 0;
      if (n != null)
      {
         if (n.getLock().isLocked())
         {
            num++;
         }
         for (Object cn : n.getChildrenDirect(true))
         {
            num += numLocks((NodeSPI) cn);
         }
      }
      return num;
   }

   /**
    * Returns an <em>approximation</em> of the total number of nodes in the
    * cache. Since this method doesn't acquire any locks, the number might be
    * incorrect, or the method might even throw a
    * ConcurrentModificationException
    */
   public int getNumberOfNodes()
   {
      return numNodes(root) - 1;
   }

   private int numNodes(NodeSPI n)
   {
      int count = 1;// for n
      if (n != null)
      {
         for (Object child : n.getChildrenDirect())
         {
            count += numNodes((NodeSPI) child);
         }
      }
      return count;
   }

   /**
    * Prints information about the contents of the nodes in the cache's current
    * in-memory state.  Does not load any previously evicted nodes from a
    * cache loader, so evicted nodes will not be included.
    */
   public String printDetails()
   {
      StringBuffer sb = new StringBuffer();
      root.printDetails(sb, 0);
      sb.append("\n");
      return sb.toString();
   }


   /**
    * Returns lock information.
    */
   public String printLockInfo()
   {
      StringBuffer sb = new StringBuffer("\n");
      int indent = 0;

      for (Object n : root.getChildrenDirect())
      {
         ((NodeSPI) n).getLock().printLockInfo(sb, indent);
         sb.append("\n");
      }
      return sb.toString();
   }

   /**
    * Returns an <em>approximation</em> of the total number of attributes in
    * this sub cache.
    *
    * @see #getNumberOfAttributes
    */
   public int getNumberOfAttributes(Fqn fqn)
   {
      return numAttributes(peek(fqn));
   }

   private int numAttributes(NodeSPI n)
   {
      int count = 0;
      for (Object child : n.getChildrenDirect())
      {
         count += numAttributes((NodeSPI) child);
      }
      count += n.getDataDirect().size();
      return count;
   }

   /**
    * Returns an <em>approximation</em> of the total number of attributes in
    * the cache. Since this method doesn't acquire any locks, the number might
    * be incorrect, or the method might even throw a
    * ConcurrentModificationException
    */
   public int getNumberOfAttributes()
   {
      return numAttributes(root);
   }

   /**
    * Removes the actual node from the tree data structure.
    *
    * @param f               the Fqn of the node to remove
    * @param skipMarkerCheck if true, skips checking the boolean {@link org.jboss.cache.NodeSPI#isDeleted()} flag and deletes the node anyway.
    * @return Returns true if the node was found and removed, false if not.
    */
   public boolean removeFromDataStructure(Fqn f, boolean skipMarkerCheck)
   {
      NodeSPI n = peek(f, true);
      if (n == null)
      {
         return false;
      }


      if (trace) log.trace("Performing a real remove for node " + f + ", marked for removal.");
      if (skipMarkerCheck || n.isDeleted())
      {
         if (n.getFqn().isRoot())
         {
            // do not actually delete; just remove deletion marker
            n.markAsDeleted(true);
            // but now remove all children, since the call has been to remove("/")
            n.removeChildrenDirect();

            // mark the node to be removed (and all children) as invalid so anyone holding a direct reference to it will
            // be aware that it is no longer valid.
            n.setValid(false, true);
            n.setValid(true, false);
            return true;
         }
         else
         {
            // mark the node to be removed (and all children) as invalid so anyone holding a direct reference to it will
            // be aware that it is no longer valid.
            n.setValid(false, true);
            return n.getParent().removeChildDirect(n.getFqn().getLastElement());
         }
      }
      else
      {
         if (log.isDebugEnabled()) log.debug("Node " + f + " NOT marked for removal as expected, not removing!");
         return false;
      }
   }

   public void evict(Fqn fqn, boolean recursive)
   {
      List<Fqn> toEvict = getNodesForEviction(fqn, recursive);
      for (Fqn aFqn : toEvict)
      {
         evict(aFqn);
      }
   }

   /**
    * @return true if the FQN is leaf and was removed; false if is an intermediate FQN and only contained data
    *         is droped.
    */
   public boolean evict(Fqn fqn)
   {
      if (peek(fqn, false, true) == null) return true;
      if (hasChildren(fqn))
      {
         if (trace)
            log.trace("removing DATA as node has children: evict(" + fqn + ")");
         removeData(fqn);
         return false;
      }
      else
      {
         if (trace)
            log.trace("removing NODE as it is a leaf: evict(" + fqn + ")");
         removeNode(fqn);
         return true;
      }
   }

   private void removeNode(Fqn fqn)
   {
      NodeSPI targetNode = peek(fqn, null, true);
      if (targetNode == null) return;
      NodeSPI parentNode = targetNode.getParent();
      if (parentNode != null)
      {
         parentNode.removeChildDirect(fqn.getLastElement());
         parentNode.setChildrenLoaded(false);
      }
   }

   protected void removeData(Fqn fqn)
   {
      NodeSPI n = peek(fqn, null);
      if (n == null)
      {
         log.warn("node " + fqn + " not found");
         return;
      }
      n.clearDataDirect();
      n.setDataLoaded(false);
   }

   /**
    * Traverses the tree to the given Fqn, creating nodes if needed.  Returns a list of nodes created, as well as a reference to the last node.
    * <p/>
    * E.g.,
    * <code>
    * Object[] results = createNode(myFqn);
    * results[0] // this is a List&lt;NodeSPI&gt; of nodes <i>created</i> in getting to the target node.
    * results[1] // is a NodeSPI reference to the target node, regardless of whether it was <i>created</i> or just <i>found</i>.
    * </code>
    *
    * @param fqn fqn to find
    * @return see above.
    */
   public Object[] createNodes(Fqn fqn)
   {
      List<NodeSPI> result = new ArrayList<NodeSPI>(fqn.size());
      Fqn tmpFqn = Fqn.ROOT;

      int size = fqn.size();

      // root node
      NodeSPI n = root;
      for (int i = 0; i < size; i++)
      {
         Object childName = fqn.get(i);
         tmpFqn = Fqn.fromRelativeElements(tmpFqn, childName);

         NodeSPI childNode;
         Map children = n.getChildrenMapDirect();
         childNode = children == null ? null : (NodeSPI) children.get(childName);

         if (childNode == null)
         {
            childNode = n.addChildDirect(Fqn.fromElements(childName));
            result.add(childNode);
         }

         n = childNode;
      }
      return new Object[]{result, n};
   }

   public void createNodesLocally(Fqn<?> fqn, Map<?, ?> data) throws CacheException
   {
      int treeNodeSize;
      if ((treeNodeSize = fqn.size()) == 0) return;
      NodeSPI n = root;
      for (int i = 0; i < treeNodeSize; i++)
      {
         Object childName = fqn.get(i);
         NodeSPI childNode = n.addChildDirect(Fqn.fromElements(childName));
         if (childNode == null)
         {
            if (trace)
            {
               log.trace("failed to find or create child " + childName + " of node " + n.getFqn());
            }
            return;
         }
         if (i == treeNodeSize - 1)
         {
            // set data
            childNode.putAllDirect(data);
         }
         n = childNode;
      }
   }
}
