package org.jboss.cache;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.buddyreplication.BuddyManager;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.factories.ComponentRegistry;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.interceptors.InterceptorChain;
import org.jboss.cache.invocation.NodeInvocationDelegate;
import org.jboss.cache.loader.CacheLoaderManager;
import org.jboss.cache.lock.LockStrategyFactory;
import org.jboss.cache.marshall.Marshaller;
import org.jboss.cache.notifications.Notifier;

import javax.management.MBeanServerFactory;
import javax.transaction.TransactionManager;
import java.util.ArrayList;

/**
 * Added in 2.2.0 to handle the lifecycle of the cache.  I.e., to control the starting and stopping process, as defined
 * by the {@link org.jboss.cache.Cache#start()} and {@link org.jboss.cache.Cache#stop()} API methods.
 * <p/>
 * This class also acts as a container for the {@link ComponentRegistry}, which it constructs in its constructor
 * as a place to hold all components of a given cache instance.
 * <p/>
 * It also holds the status of the cache, which can be queried.
 * <p/>
 *
 * @author Mircea.Markus@jboss.com
 * @since 2.2
 */
public class LifecycleManager
{
   private static final Log log = LogFactory.getLog(LifecycleManager.class);
   private CacheStatus cacheStatus;

   /* dependencies*/
   private Configuration configuration;
   private Notifier notifier;
   private RegionManager regionManager;
   private NodeFactory nodeFactory;
   private DataContainer dataContainer;
   private BuddyManager buddyManager;
   private RPCManager rpcManager;
   private ComponentRegistry componentRegistry;


   @Inject
   public void initialize(Configuration configuration, Notifier notifier, RegionManager regionManager, NodeFactory nodeFactory,
                          DataContainer cacheData, BuddyManager buddyManager, RPCManager rpcManager,
                          ComponentRegistry componentRegistry)
   {
      this.configuration = configuration;
      this.notifier = notifier;
      this.regionManager = regionManager;
      this.nodeFactory = nodeFactory;
      this.buddyManager = buddyManager;
      this.rpcManager = rpcManager;
      this.dataContainer = cacheData;
      this.componentRegistry = componentRegistry;
   }

   /**
    * Hook to shut down the cache when the JVM exits.
    */
   private Thread shutdownHook;
   /**
    * A flag that the shutdown hook sets before calling cache.stop().  Allows stop() to identify if it has been called
    * from a shutdown hook.
    */
   private boolean invokedFromShutdownHook;

   /**
    * Constructs a new instance, also constructs a {@link org.jboss.cache.factories.ComponentRegistry} instance which can
    * then be retrieved using {@link #getComponentRegistry()}.  When constructed, the cache status is set to {@link org.jboss.cache.CacheStatus#INSTANTIATED}.
    *
    * @param configuration with which to create this class.
    */
   public LifecycleManager(Configuration configuration)
   {
      this.configuration = configuration;
      this.componentRegistry = new ComponentRegistry(configuration);
      this.cacheStatus = CacheStatus.INSTANTIATED;
   }

   /**
    * Creates the components needed by a cache instance and sets the cache status to {@link org.jboss.cache.CacheStatus#CREATED}
    * when it is done.
    *
    * @throws CacheException if there is a problem with construction.
    */
   public void create() throws CacheException
   {
      if (!cacheStatus.createAllowed())
      {
         if (cacheStatus.needToDestroyFailedCache())
            destroy();
         else
            return;
      }
      try
      {
         internalCreate();
      }
      catch (Throwable t)
      {
         handleLifecycleTransitionFailure(t);
      }
   }

   /**
    * Starts the components needed by a cache instance, and then starts the cache instance itself.  Sets the cache status to
    * {@link org.jboss.cache.CacheStatus#STARTED} once it is done.
    * <p/>
    * If the cache status is not {@link org.jboss.cache.CacheStatus#CREATED} when this is called, it will first call {@link #create()}
    * to create the cache.
    * <p/>
    *
    * @throws CacheException if there is a problem with starting the cache.
    */
   public void start() throws CacheException
   {
      if (!cacheStatus.startAllowed())
      {
         if (cacheStatus.needToDestroyFailedCache())
            destroy(); // this will take us back to DESTROYED

         if (cacheStatus.needCreateBeforeStart())
            create();
         else
            return;
      }

      try
      {
         internalStart();
      }
      catch (Throwable t)
      {
//         if (log.isTraceEnabled()) log.trace("InternalStart had problems: ", t);
         handleLifecycleTransitionFailure(t);
      }
   }

   /**
    * Destroys the cache and frees up any resources.  Sets the cache status to {@link CacheStatus#DESTROYED} when it is done.
    * <p/>
    * If the cache is in {@link org.jboss.cache.CacheStatus#STARTED} when this method is called, it will first call {@link #stop()}
    * to stop the cache.
    */
   public void destroy()
   {
      if (!cacheStatus.destroyAllowed())
      {
         if (cacheStatus.needStopBeforeDestroy())
         {
            try
            {
               stop();
            }
            catch (CacheException e)
            {
               log.warn("Needed to call stop() before destroying but stop() " +
                     "threw exception. Proceeding to destroy", e);
            }
         }
         else
            return;
      }

      try
      {
         internalDestroy();
      }
      finally
      {
         // We always progress to destroyed
         cacheStatus = CacheStatus.DESTROYED;
      }
   }

   /**
    * Stops the cache and sets the cache status to {@link org.jboss.cache.CacheStatus#STOPPED} once it is done.  If the cache is not in
    * the {@link org.jboss.cache.CacheStatus#STARTED} state, this is a no-op.
    */
   public void stop()
   {
      if (!cacheStatus.stopAllowed())
      {
         return;
      }

      // Trying to stop() from FAILED is valid, but may not work
      boolean failed = cacheStatus == CacheStatus.FAILED;

      try
      {
         internalStop();
      }
      catch (Throwable t)
      {
         if (failed)
         {
            log.warn("Attempted to stop() from FAILED state, " +
                  "but caught exception; try calling destroy()", t);
         }
         handleLifecycleTransitionFailure(t);
      }
      finally
      {
         if (!failed) cacheStatus = CacheStatus.STOPPED;
      }
   }

   /**
    * The actual start implementation.
    */
   private void internalStart() throws CacheException, IllegalArgumentException
   {
      // re-wire all dependencies in case stuff has changed since the cache was created

      // remove any components whose construction may have depended upon a configuration that may have changed.
      removeConfigurationDependentComponents();

      // this will recreate any missing components based on the current config
      componentRegistry.updateDependencies();
      componentRegistry.wireDependencies(dataContainer.getRoot());

      cacheStatus = CacheStatus.STARTING;

      // start all internal components
      componentRegistry.start();

      if (log.isDebugEnabled())
         log.debug("Interceptor chain is: " + componentRegistry.getComponent(InterceptorChain.class));

      correctRootNodeType();

      startManualComponents();

      // start any eviction threads.
      if (regionManager.isUsingEvictions())
      {
         regionManager.startEvictionThread();
      }

      notifier.notifyCacheStarted();

      addShutdownHook();

      log.info("JBoss Cache version: " + Version.printVersion());

      cacheStatus = CacheStatus.STARTED;
   }

   private void addShutdownHook()
   {
      ArrayList al = MBeanServerFactory.findMBeanServer(null);
      boolean registerShutdownHook = (configuration.getShutdownHookBehavior() == Configuration.ShutdownHookBehavior.DEFAULT && al.size() == 0)
            || configuration.getShutdownHookBehavior() == Configuration.ShutdownHookBehavior.REGISTER;

      if (registerShutdownHook)
      {
         if (log.isTraceEnabled())
            log.trace("Registering a shutdown hook.  Configured behavior = " + configuration.getShutdownHookBehavior());
         shutdownHook = new Thread()
         {
            @Override
            public void run()
            {
               try
               {
                  invokedFromShutdownHook = true;
                  stop();
               }
               finally
               {
                  invokedFromShutdownHook = false;
               }
            }
         };

         Runtime.getRuntime().addShutdownHook(shutdownHook);
      }
      else
      {
         if (log.isTraceEnabled())
            log.trace("Not registering a shutdown hook.  Configured behavior = " + configuration.getShutdownHookBehavior());
      }
   }

   /**
    * The actual create implementation.
    *
    * @throws CacheException
    */
   private void internalCreate() throws CacheException
   {
      componentRegistry.wire();
      correctRootNodeType();

      LockStrategyFactory.setIsolationLevel(configuration.getIsolationLevel());

      cacheStatus = CacheStatus.CREATED;
   }

   /**
    * Sets the cacheStatus to FAILED and rethrows the problem as one
    * of the declared types. Converts any non-RuntimeException Exception
    * to CacheException.
    *
    * @param t
    * @throws CacheException
    * @throws RuntimeException
    * @throws Error
    */
   private void handleLifecycleTransitionFailure(Throwable t)
         throws RuntimeException, Error
   {
      cacheStatus = CacheStatus.FAILED;
      if (t instanceof CacheException)
         throw (CacheException) t;
      else if (t instanceof RuntimeException)
         throw (RuntimeException) t;
      else if (t instanceof Error)
         throw (Error) t;
      else
         throw new CacheException(t);
   }

   /**
    * The actual destroy implementation.
    */
   private void internalDestroy()
   {
      cacheStatus = CacheStatus.DESTROYING;
      // The rest of these should have already been taken care of in stop,
      // but we do it here as well in case stop failed.
      rpcManager.stop();
      componentRegistry.reset();
   }

   /**
    * The actual stop implementation.
    */
   private void internalStop()
   {
      cacheStatus = CacheStatus.STOPPING;
      // if this is called from a source other than the shutdown hook, deregister the shutdown hook.
      if (!invokedFromShutdownHook && shutdownHook != null) Runtime.getRuntime().removeShutdownHook(shutdownHook);
      componentRegistry.stop();
      if (notifier != null)
      {
         notifier.notifyCacheStopped();
         notifier.removeAllCacheListeners();
      }
      // unset transaction manager reference
      cacheStatus = CacheStatus.STOPPED;
      // empty in-memory state
      dataContainer.getRoot().clearDataDirect();
      dataContainer.getRoot().removeChildrenDirect();
   }

   private void removeConfigurationDependentComponents()
   {
      // remove the Interceptor.class component though, since it may pertain to an old config

      componentRegistry.unregisterComponent(InterceptorChain.class); // the interceptor chain will need reconstructing in case the cfg has changed before calling start()
      componentRegistry.unregisterComponent(Marshaller.class);
      componentRegistry.unregisterComponent(TransactionManager.class);
      componentRegistry.unregisterComponent(BuddyManager.class);
      componentRegistry.unregisterComponent(CacheLoaderManager.class);
   }

   /**
    * Creates a new root node if one does not exist, or if the existing one does not match the type according to the configuration.
    */
   private void correctRootNodeType()
   {
      // create a new root temporarily.
      NodeSPI tempRoot = nodeFactory.createRootDataNode();
      // if we don't already have a root or the new (temp) root is of a different class (optimistic vs pessimistic) to
      // the current root, then we use the new one.

      Class currentRootType = dataContainer.getRoot() == null ? null : ((NodeInvocationDelegate) dataContainer.getRoot()).getDelegationTarget().getClass();
      Class tempRootType = ((NodeInvocationDelegate) tempRoot).getDelegationTarget().getClass();

      if (!tempRootType.equals(currentRootType)) dataContainer.setRoot(tempRoot);
   }

   private void startManualComponents()
   {
      // these 2 components need to be started manually since they can only be started after ALL other components have started.
      // i.e., rpcManager's start() method may do state transfers.  State transfers will rely on the interceptor chain being started.
      // the interceptor chain cannot start until the rpcManager is started.  And similarly, the buddyManager relies on the
      // rpcManager being started.
      if (rpcManager != null) rpcManager.start();
      if (buddyManager != null)
      {
         buddyManager.init();
         if (buddyManager.isEnabled())
         {
            dataContainer.registerInternalFqn(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN);
         }
      }
   }

   /**
    * Asserts whether invocations are allowed on the cache or not.  Returns <tt>true</tt> if invocations are to be allowed,
    * <tt>false</tt> otherwise.  If the origin of the call is remote and the cache status is {@link org.jboss.cache.CacheStatus#STARTING},
    * this method will block for up to {@link org.jboss.cache.config.Configuration#getStateRetrievalTimeout()} millis, checking
    * for a valid state.
    *
    * @param originLocal true if the call originates locally (i.e., from the {@link org.jboss.cache.invocation.CacheInvocationDelegate} or false if it originates remotely, i.e., from the {@link org.jboss.cache.marshall.CommandAwareRpcDispatcher}.
    * @return true if invocations are allowed, false otherwise.
    */
   public boolean invocationsAllowed(boolean originLocal)
   {
      if (getCacheStatus().allowInvocations()) return true;

      // if this is a locally originating call and the cache is not in a valid state, return false.
      if (originLocal) return false;

      // else if this is a remote call and the status is STARTING, wait until the cache starts.
      if (getCacheStatus() == CacheStatus.STARTING)
      {
         try
         {
            blockUntilCacheStarts();
            return true;
         }
         catch (InterruptedException e)
         {
            Thread.currentThread().interrupt();
         }
      }
      else
      {
         log.warn("Received a remote call but the cache is not in STARTED state - ignoring call.");
      }
      return false;
   }

   /**
    * Blocks until the current cache instance is in it's {@link org.jboss.cache.CacheStatus#STARTED started} phase.  Blocks
    * for up to {@link org.jboss.cache.config.Configuration#getStateRetrievalTimeout()} milliseconds, throwing an IllegalStateException
    * if the cache doesn't reach this state even after this maximum wait time.
    *
    * @throws InterruptedException  if interrupted while waiting
    * @throws IllegalStateException if even after waiting the cache has not started.
    */
   private void blockUntilCacheStarts() throws InterruptedException, IllegalStateException
   {
      int pollFrequencyMS = 100;
      long startupWaitTime = configuration.getStateRetrievalTimeout();
      long giveUpTime = System.currentTimeMillis() + startupWaitTime;

      while (System.currentTimeMillis() < giveUpTime)
      {
         if (getCacheStatus().allowInvocations()) break;
         Thread.sleep(pollFrequencyMS);
      }

      // check if we have started.
      if (!getCacheStatus().allowInvocations())
         throw new IllegalStateException("Cache not in STARTED state, even after waiting " + configuration.getStateRetrievalTimeout() + " millis.");
   }

   public CacheStatus getCacheStatus()
   {
      return cacheStatus;
   }

   public ComponentRegistry getComponentRegistry()
   {
      return componentRegistry;
   }

   public void setCacheStatus(CacheStatus cacheStatus)
   {
      this.cacheStatus = cacheStatus;
   }
}
