package org.jboss.cache.commands.write;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheException;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.Node;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.commands.Visitor;
import org.jboss.cache.config.Option;
import org.jboss.cache.optimistic.DataVersion;

import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import java.util.Collections;
import java.util.Map;

/**
 * Very much like an evict(), except that regardless of whether there is a child present, this call will never
 * remove the node from memory - just remove its contents.
 * <p/>
 * Also, potentially throws a cache exception if data versioning is used and the node in memory has a newer data
 * version than what is passed in.
 * <p/>
 * Finally, the data version of the in-memory node is updated to the version being evicted to prevent versions
 * going out of sync.
 *
 * @author Mircea.Markus@jboss.com
 * @since 2.2
 */
public class InvalidateCommand extends EvictCommand
{
   public static final int METHOD_ID = 47;
   private static final Log log = LogFactory.getLog(InvalidateCommand.class);
   private static boolean trace = log.isTraceEnabled();

   /* dependencies*/
   private boolean isOptimisticLocking;
   private TransactionManager transactionManager;
   private CacheSPI spi;

   public InvalidateCommand(Fqn fqn)
   {
      super(fqn);
   }

   public InvalidateCommand()
   {
   }

   public void initialize(TransactionManager txManager, CacheSPI cacheSpi, boolean optimistic)
   {
      this.transactionManager = txManager;
      this.spi = cacheSpi;
      this.isOptimisticLocking = optimistic;
   }

   /**
    * Performs an invalidation on a specified node
    *
    * @param ctx invocation context
    * @return null
    */
   @Override
   public Object perform(InvocationContext ctx)
   {
      //TODO: 2.2.0: rather than using CacheSPI this should use peek().  The other interceptors should obtain locks and load nodes if necessary for this InvalidateCommand.
      Node node = spi.getNode(fqn); // force interceptor chain, load if necessary from cache loader.

      if (trace) log.trace("Invalidating fqn:" + fqn);
      if (node == null)
      {
         // if pessimistic locking, just return.
         if (!isOptimisticLocking) return null;

         // check if a tombstone already exists
         NodeSPI nodeSPI = dataContainer.peek(fqn, false, true);
         if (nodeSPI == null)
         {
            if (dataVersion == null)
            {
               if (trace)
                  log.trace("Would have created a tombstone since the node doesn't exist, but the version to invalidate is null and hence cannot create a tombstone!");
               return null;
            }
            if (trace)
               log.trace("Node doesn't exist; creating a tombstone with data version " + dataVersion);
            // create the node we need.
            Map m = Collections.emptyMap();
            InvocationContext ic = spi.getInvocationContext();
            Option o = ic.getOptionOverrides();
            boolean origCacheModeLocal = o.isCacheModeLocal();
            o.setCacheModeLocal(true);
            o.setDataVersion(dataVersion);
            // if we are in a tx this call should happen outside of any tx
            try
            {
               Transaction suspended = null;
               if (transactionManager != null)
               {
                  suspended = transactionManager.suspend();
               }
               spi.put(fqn, m);
               if (suspended != null) transactionManager.resume(suspended);
               ic.getOptionOverrides().setCacheModeLocal(origCacheModeLocal);
            }
            catch (Exception e)
            {
               log.error("Unable to create tombstone!", e);
            }
            nodeSPI = (NodeSPI) dataContainer.getRoot().getChild(fqn);
         }
         node = nodeSPI;
      }

      if (isOptimisticLocking)
         removeData();
      else
         super.perform(ctx);

      // mark the node to be removed (and all children) as invalid so anyone holding a direct reference to it will
      // be aware that it is no longer valid.
      NodeSPI nSPI = (NodeSPI) node;
      nSPI.setValid(false, true);
      // root nodes can never be invalid
      if (fqn.isRoot()) nSPI.setValid(true, false); // non-recursive.

      if (dataVersion != null)
      {
         NodeSPI n = dataContainer.peek(fqn, false, true);
         n.setVersion(dataVersion);
      }
      return null;
   }

   private void removeData()
         throws CacheException
   {
      // Find the node. This will lock it (if <tt>locking</tt> is true) and
      // add the temporarily created parent nodes to the TX's node list if tx != null)
      NodeSPI n = dataContainer.peek(fqn, dataVersion);
      if (n == null)
      {
         log.warn("node " + fqn + " not found");
         return;
      }
      InvocationContext ctx = spi.getInvocationContext();
      notifier.notifyNodeEvicted(fqn, true, ctx);
      n.clearDataDirect();
      n.setDataLoaded(false);
      notifier.notifyNodeEvicted(fqn, false, ctx);
   }

   @Override
   public Object acceptVisitor(InvocationContext ctx, Visitor visitor) throws Throwable
   {
      return visitor.visitInvalidateCommand(ctx, this);
   }

   @Override
   public boolean isVersioned()
   {
      // invalidate commands are *always* versioned.
      return true;
   }

   @Override
   protected boolean isVersionedId(int id)
   {
      // invalidate commands are *always* versioned.
      return true;
   }

   @Override
   public int getCommandId()
   {
      return METHOD_ID;
   }

   @Override
   public String toString()
   {
      return "InvalidateCommand{" +
            "fqn=" + fqn +
            ", dataVersion=" + dataVersion +
            '}';
   }

   @Override
   public Object[] getParameters()
   {
      return new Object[]{fqn, dataVersion};
   }

   @Override
   public void setParameters(int commandId, Object[] args)
   {
      fqn = (Fqn) args[0];
      dataVersion = (DataVersion) args[1];
   }

}
