package org.jboss.cache.commands.write;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.commands.Visitor;
import org.jboss.cache.notifications.event.NodeModifiedEvent;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.transaction.GlobalTransaction;

import java.util.HashMap;
import java.util.Map;

/**
 * Implements functionality defined by {@link org.jboss.cache.Cache#remove(String, Object)}
 *
 * @author Mircea.Markus@jboss.com
 * @since 2.2
 */
public class RemoveDataCommand extends AbstractVersionedDataCommand
{
   public static final int METHOD_ID = 7;
   public static final int VERSIONED_METHOD_ID = 42;

   private static final Log log = LogFactory.getLog(RemoveDataCommand.class);
   private static boolean trace = log.isTraceEnabled();

   /* parameters*/
   private boolean createUndoops;
   private boolean sendNodeEvent;
   private boolean eviction;
   private HashMap originalData;

   public RemoveDataCommand(GlobalTransaction gtx, Fqn fqn, boolean createUndoops, boolean sendNodeEvent, boolean eviction)
   {
      this.globalTransaction = gtx;
      this.fqn = fqn;
      this.createUndoops = createUndoops;
      this.sendNodeEvent = sendNodeEvent;
      this.eviction = eviction;
   }

   public RemoveDataCommand()
   {
   }

   /**
    * Clears the data map in the node referenced by the specified Fqn.
    *
    * @param ctx invocation context
    * @return null
    */
   public Object perform(InvocationContext ctx)
   {
      if (trace) log.trace("perform(" + globalTransaction + ", \"" + fqn + "\")");
      NodeSPI targetNode = dataContainer.peek(fqn, dataVersion);
      if (targetNode == null)
      {
         log.warn("node " + fqn + " not found");
         return null;
      }

      Map data = targetNode.getDataDirect();
      prepareDataForRollback(data);

      notifyBefore(data, ctx);

      targetNode.clearDataDirect();
      if (eviction) targetNode.setDataLoaded(false);

      notifyAfter(data, ctx);

      return null;
   }

   private void prepareDataForRollback(Map data)
   {
      if (globalTransaction != null && createUndoops && !eviction && !data.isEmpty())
      {
         originalData = new HashMap(data);
      }
   }

   private void notifyAfter(Map data, InvocationContext ctx)
   {
      if (sendNodeEvent)
      {
         notifier.notifyNodeVisited(fqn, false, ctx);
      }
      else
      {
         if (eviction)
         {
            notifier.notifyNodeEvicted(fqn, false, ctx);
         }
         else
         {
            notifier.notifyNodeModified(fqn, false, NodeModifiedEvent.ModificationType.REMOVE_DATA, data, ctx);
         }
      }
   }

   private void notifyBefore(Map data, InvocationContext ctx)
   {
      if (eviction)
      {
         notifier.notifyNodeEvicted(fqn, true, ctx);
      }
      else
      {
         notifier.notifyNodeModified(fqn, true, NodeModifiedEvent.ModificationType.REMOVE_DATA, data, ctx);
      }
   }

   public void rollback()
   {
      if (trace) log.trace("rollback(" + globalTransaction + ", \"" + fqn + "\", " + originalData + ")");
      boolean isRollback = true;
      NodeSPI nodeSPI = dataContainer.peekStrict(globalTransaction, fqn, isRollback);
      nodeSPI.putAllDirect(originalData);
   }

   public Object acceptVisitor(InvocationContext ctx, Visitor visitor) throws Throwable
   {
      return visitor.visitRemoveDataCommand(ctx, this);
   }

   public boolean isEviction()
   {
      return eviction;
   }

   public boolean isSendNodeEvent()
   {
      return sendNodeEvent;
   }

   public boolean isCreateUndoops()
   {
      return createUndoops;
   }

   public int getCommandId()
   {
      return isVersioned() ? VERSIONED_METHOD_ID : METHOD_ID;
   }

   @Override
   public Object[] getParameters()
   {
      if (isVersioned())
         return new Object[]{globalTransaction, fqn, createUndoops, dataVersion};
      else
         return new Object[]{globalTransaction, fqn, createUndoops};
   }

   @Override
   public void setParameters(int commandId, Object[] args)
   {
      globalTransaction = (GlobalTransaction) args[0];
      fqn = (Fqn) args[1];
      createUndoops = (Boolean) args[2];
      if (isVersionedId(commandId)) dataVersion = (DataVersion) args[3];
   }

   @Override
   public boolean equals(Object o)
   {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;
      if (!super.equals(o)) return false;

      RemoveDataCommand that = (RemoveDataCommand) o;

      if (createUndoops != that.createUndoops) return false;
      if (eviction != that.eviction) return false;
      if (sendNodeEvent != that.sendNodeEvent) return false;
      if (globalTransaction != null ? !globalTransaction.equals(that.globalTransaction) : that.globalTransaction != null)
         return false;

      return true;
   }

   @Override
   public int hashCode()
   {
      int result = super.hashCode();
      result = 31 * result + (globalTransaction != null ? globalTransaction.hashCode() : 0);
      result = 31 * result + (createUndoops ? 1 : 0);
      result = 31 * result + (sendNodeEvent ? 1 : 0);
      result = 31 * result + (eviction ? 1 : 0);
      return result;
   }

   @Override
   protected boolean isVersionedId(int id)
   {
      return id == VERSIONED_METHOD_ID;
   }

   @Override
   public String toString()
   {
      return "RemoveDataCommand{" +
            "fqn=" + fqn +
            ", dataVersion=" + dataVersion +
            ", globalTransaction=" + globalTransaction +
            ", createUndoops=" + createUndoops +
            ", sendNodeEvent=" + sendNodeEvent +
            ", eviction=" + eviction +
            ", originalData=" + originalData +
            '}';
   }
}
