/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheException;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.commands.VisitableCommand;
import org.jboss.cache.commands.tx.CommitCommand;
import org.jboss.cache.commands.tx.OptimisticPrepareCommand;
import org.jboss.cache.commands.tx.RollbackCommand;
import org.jboss.cache.factories.annotations.Start;
import static org.jboss.cache.lock.NodeLock.LockType.READ;
import static org.jboss.cache.lock.NodeLock.LockType.WRITE;
import org.jboss.cache.optimistic.TransactionWorkspace;
import org.jboss.cache.optimistic.WorkspaceNode;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionEntry;

/**
 * Locks nodes during transaction boundaries.  Only affects prepare/commit/rollback method calls; other method calls
 * are simply passed up the interceptor stack.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 * @author <a href="mailto:stevew@jofti.com">Steve Woodcock (stevew@jofti.com)</a>
 */
public class OptimisticLockingInterceptor extends OptimisticInterceptor
{
   private long lockAcquisitionTimeout;

   @Start
   private void init()
   {
      if (txManager == null)
         log.fatal("No transaction manager lookup class has been defined. Transactions cannot be used and thus OPTIMISTIC locking cannot be used!  Expect errors!!");
      lockAcquisitionTimeout = configuration.getLockAcquisitionTimeout();
   }

   @Override
   public Object visitOptimisticPrepareCommand(InvocationContext ctx, OptimisticPrepareCommand command) throws Throwable
   {
      //try and acquire the locks - before passing on
      GlobalTransaction gtx = getGlobalTransaction(ctx);
      long timeout = lockAcquisitionTimeout;
      if (ctx.getOptionOverrides() != null
            && ctx.getOptionOverrides().getLockAcquisitionTimeout() >= 0)
      {
         timeout = ctx.getOptionOverrides().getLockAcquisitionTimeout();
      }
      try
      {
         TransactionWorkspace<?, ?> workspace = getTransactionWorkspace(gtx);
         TransactionEntry te = txTable.get(gtx);
         if (log.isDebugEnabled()) log.debug("Locking nodes in transaction workspace for GlobalTransaction " + gtx);

         for (WorkspaceNode workspaceNode : workspace.getNodes().values())
         {
            NodeSPI node = workspaceNode.getNode();

            boolean isWriteLockNeeded = workspaceNode.isDirty() || (workspaceNode.isChildrenModified() && (configuration.isLockParentForChildInsertRemove() || node.isLockForChildInsertRemove()));

            boolean acquired = node.getLock().acquire(gtx, timeout, isWriteLockNeeded ? WRITE : READ);
            if (acquired)
            {
               if (trace) log.trace("Acquired lock on node " + node.getFqn());
               te.addLock(node.getLock());
            }
            else
            {
               throw new CacheException("Unable to acquire lock on node " + node.getFqn());
            }

         }
      }
      catch (Throwable e)
      {
         log.debug("Caught exception attempting to lock nodes ", e);
         //we have failed - set to rollback and throw exception
         try
         {
            unlock(gtx);
         }
         catch (Throwable t)
         {
            // we have failed to unlock - now what?
            log.error("Failed to unlock nodes, after failing to lock nodes during a prepare!  Locks are possibly in a very inconsistent state now!", t);
         }
         throw e;
      }

      // locks have acquired so lets pass on up
      return invokeNextInterceptor(ctx, command);
   }

   @Override
   public Object visitCommitCommand(InvocationContext ctx, CommitCommand command) throws Throwable
   {
      return transactionFinalized(ctx, command);
   }

   @Override
   public Object visitRollbackCommand(InvocationContext ctx, RollbackCommand command) throws Throwable
   {
      return transactionFinalized(ctx, command);
   }

   private Object transactionFinalized(InvocationContext ctx, VisitableCommand command) throws Throwable
   {
      Object retval = null;
      // we need to let the stack run its commits or rollbacks first -
      // we unlock last - even if an exception occurs
      try
      {
         retval = invokeNextInterceptor(ctx, command);
      }
      finally
      {
         try
         {
            unlock(getGlobalTransaction(ctx));
         }
         catch (Exception e)
         {
            // we have failed to unlock - now what?
            log.error("Failed to unlock nodes after a commit or rollback!  Locks are possibly in a very inconsistent state now!", e);
         }
      }
      return retval;
   }

   /**
    * Releases all locks held by the specified global transaction.
    *
    * @param gtx which holds locks
    */
   private void unlock(GlobalTransaction gtx)
   {
      TransactionEntry entry = txTable.get(gtx);
      entry.releaseAllLocksFIFO(gtx);
   }

}
