/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.interceptors;

import org.jboss.cache.CacheException;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.commands.AbstractVisitor;
import org.jboss.cache.commands.ReplicableCommand;
import org.jboss.cache.commands.ReversibleCommand;
import org.jboss.cache.commands.VersionedDataCommand;
import org.jboss.cache.commands.VisitableCommand;
import org.jboss.cache.commands.read.GravitateDataCommand;
import org.jboss.cache.commands.tx.CommitCommand;
import org.jboss.cache.commands.tx.OptimisticPrepareCommand;
import org.jboss.cache.commands.tx.RollbackCommand;
import org.jboss.cache.commands.write.CreateNodeCommand;
import org.jboss.cache.commands.write.EvictCommand;
import org.jboss.cache.commands.write.MoveCommand;
import org.jboss.cache.commands.write.PutDataMapCommand;
import org.jboss.cache.commands.write.PutKeyValueCommand;
import org.jboss.cache.commands.write.RemoveDataCommand;
import org.jboss.cache.commands.write.RemoveKeyCommand;
import org.jboss.cache.commands.write.RemoveNodeCommand;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.Option;
import org.jboss.cache.factories.CommandsFactory;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.optimistic.DataVersion;
import org.jboss.cache.optimistic.DefaultDataVersion;
import org.jboss.cache.optimistic.TransactionWorkspace;
import org.jboss.cache.optimistic.WorkspaceNode;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.OptimisticTransactionEntry;
import org.jboss.cache.util.concurrent.ConcurrentHashSet;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

/**
 * Replication interceptor for the optimistically locked interceptor chain.  Responsible for replicating
 * state to remote nodes.  Unlike it's cousin, the {@link org.jboss.cache.interceptors.ReplicationInterceptor}, this interceptor
 * only deals with transactional calls.  Just like all things to do with Optimistic Locking, it is a requirement that
 * everything is done in a transaction and the transaction context is available via {@link org.jboss.cache.InvocationContext#getTransaction()}
 * and {@link org.jboss.cache.InvocationContext#getGlobalTransaction()}.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 * @author <a href="mailto:stevew@jofti.com">Steve Woodcock (stevew@jofti.com)</a>
 */
public class OptimisticReplicationInterceptor extends BaseRpcInterceptor
{
   /**
    * record of local broacasts - so we do not broadcast rollbacks/commits that resuted from local prepare failures  we
    * really just need a set here, but concurrent CopyOnWriteArraySet has poor performance when writing.
    */
   private final Set<GlobalTransaction> broadcastTxs = new ConcurrentHashSet<GlobalTransaction>();

   private CommandsFactory commandsFactory;

   @Inject
   public void initialize(CommandsFactory commandsFactory)
   {
      this.commandsFactory = commandsFactory;
   }

   private boolean isLocalOptionOverrides(InvocationContext ctx)
   {
      Option optionOverride = ctx.getOptionOverrides();
      if (optionOverride != null && optionOverride.isCacheModeLocal() && ctx.getTransaction() == null)
      {
         // skip replication!!
         log.debug("Skipping replication for this call as cache mode is local, forced via an option override.");
         return true;
      }
      return false;
   }

   @Override
   public Object visitOptimisticPrepareCommand(InvocationContext ctx, OptimisticPrepareCommand command) throws Throwable
   {
      // pass up the chain.
      Object retval = invokeNextInterceptor(ctx, command);
      if (!skipReplicationOfTransactionMethod(ctx))
      {
         GlobalTransaction gtx = getGlobalTransaction(ctx);

         if (!gtx.isRemote() && ctx.isOriginLocal())
         {
            // replicate the prepare call.
            broadcastPrepare(command, gtx, ctx);
         }
      }
      return retval;
   }

   @Override
   public Object visitCommitCommand(InvocationContext ctx, CommitCommand command) throws Throwable
   {
      if (isLocalOptionOverrides(ctx))
      {
         return invokeNextInterceptor(ctx, command);
      }
      //lets broadcast the commit first
      Throwable remoteCommitException = null;
      GlobalTransaction gtx = getGlobalTransaction(ctx);
      if (!gtx.isRemote() && ctx.isOriginLocal() && broadcastTxs.contains(gtx))
      {
         //we dont do anything
         try
         {
            if (!skipReplicationOfTransactionMethod(ctx)) broadcastCommit(gtx, ctx);
         }
         catch (Throwable t)
         {
            log.error("A problem occurred with remote commit", t);
            remoteCommitException = t;
         }
      }

      Object retval = invokeNextInterceptor(ctx, command);
      if (remoteCommitException != null)
      {
         throw remoteCommitException;
      }
      return retval;
   }

   @Override
   public Object visitRollbackCommand(InvocationContext ctx, RollbackCommand command) throws Throwable
   {
      if (isLocalOptionOverrides(ctx))
      {
         return invokeNextInterceptor(ctx, command);
      }
      //    lets broadcast the rollback first
      GlobalTransaction gtx = getGlobalTransaction(ctx);
      Throwable remoteRollbackException = null;
      if (!gtx.isRemote() && ctx.isOriginLocal() && broadcastTxs.contains(gtx))
      {
         //we dont do anything
         try
         {
            if (!skipReplicationOfTransactionMethod(ctx)) broadcastRollback(gtx, ctx);
         }
         catch (Throwable t)
         {
            log.error(" a problem occurred with remote rollback", t);
            remoteRollbackException = t;
         }

      }
      Object retval = invokeNextInterceptor(ctx, command);
      if (remoteRollbackException != null)
      {
         throw remoteRollbackException;
      }
      return retval;
   }

   @Override
   public Object visitPutKeyValueCommand(InvocationContext ctx, PutKeyValueCommand command) throws Throwable
   {
      if (isLocalOptionOverrides(ctx))
      {
         return invokeNextInterceptor(ctx, command);
      }
      GlobalTransaction gtx = getGlobalTransaction(ctx);
      if (command.isPutForExternalRead()) txTable.get(gtx).setForceAsyncReplication(true);
      return invokeNextInterceptor(ctx, command);
   }


   private GlobalTransaction getGlobalTransaction(InvocationContext ctx)
   {
      // get the current globalTransaction
      GlobalTransaction gtx = ctx.getGlobalTransaction();
      if (gtx == null)
      {
         throw new CacheException("failed to get global transaction");
      }
      return gtx;
   }

   protected void broadcastPrepare(OptimisticPrepareCommand command, GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      boolean remoteCallSync = configuration.getCacheMode() == Configuration.CacheMode.REPL_SYNC;

      // this method will return immediately if we're the only member
      if (rpcManager.getMembers() != null && rpcManager.getMembers().size() > 1)
      {
         // Map method calls to data versioned equivalents.
         // See JBCACHE-843 and docs/design/DataVersioning.txt
         //TODO: 2.2.0: make the visitor a field
         DataVersionPopulator populator = new DataVersionPopulator(getTransactionWorkspace(gtx));
         List<ReversibleCommand> clonedModifications = new ArrayList<ReversibleCommand>(command.getModifications().size());
         for (ReversibleCommand command1 : command.getModifications())
         {
            ReversibleCommand clone = (ReversibleCommand) command1.acceptVisitor(null, populator);
            clonedModifications.add(clone);
         }
         ReplicableCommand toBroadcast = commandsFactory.buildOptimisticPrepareCommand(gtx, clonedModifications, command.getData(), command.getLocalAddress(), command.isOnePhaseCommit());

         //record the things we have possibly sent
         broadcastTxs.add(gtx);
         if (log.isDebugEnabled())
         {
            log.debug("(" + rpcManager.getLocalAddress() + "): broadcasting prepare for " + gtx + " (" + command.getModificationsCount() + " modifications");
         }
         replicateCall(ctx, toBroadcast, remoteCallSync, ctx.getOptionOverrides());
      }
      else
      {
         //no members, ignoring
         if (log.isDebugEnabled())
         {
            log.debug("(" + rpcManager.getLocalAddress() + "):not broadcasting prepare as members are " + rpcManager.getMembers());
         }
      }
   }


   protected void broadcastCommit(GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      boolean remoteCallSync = configuration.isSyncCommitPhase();

      // Broadcast commit() to all members (exclude myself though)
      if (rpcManager.getMembers() != null && rpcManager.getMembers().size() > 1)
      {
         try
         {
            broadcastTxs.remove(gtx);
            CommitCommand commitCommand = commandsFactory.buildCommitCommand(gtx);

            if (log.isDebugEnabled())
               log.debug("running remote commit for " + gtx + " and coord=" + rpcManager.getLocalAddress());

            // for an optimistic commit we don't need to force an OOB message since O/L means we have non-blocking reads.
            replicateCall(ctx, commitCommand, remoteCallSync, ctx.getOptionOverrides(), false);
         }
         catch (Exception e)
         {
            log.error("Commit failed", e);
            throw e;
         }
      }
   }

   protected void broadcastRollback(GlobalTransaction gtx, InvocationContext ctx) throws Throwable
   {
      boolean remoteCallSync = configuration.isSyncRollbackPhase();

      if (rpcManager.getMembers() != null && rpcManager.getMembers().size() > 1)
      {
         // Broadcast rollback() to all other members (excluding myself)
         try
         {
            broadcastTxs.remove(gtx);
            RollbackCommand rollbackCommand = commandsFactory.buildRollbackCommand(null);

            if (log.isDebugEnabled())
               log.debug("running remote rollback for " + gtx + " and coord=" + rpcManager.getLocalAddress());
            replicateCall(ctx, rollbackCommand, remoteCallSync, ctx.getOptionOverrides());
         }
         catch (Exception e)
         {
            log.error("Rollback failed", e);
            throw e;
         }
      }
   }

   public class DataVersionPopulator extends AbstractVisitor
   {
      final TransactionWorkspace workspace;

      public DataVersionPopulator(TransactionWorkspace workspace)
      {
         this.workspace = workspace;
      }

      @Override
      public Object visitGravitateDataCommand(InvocationContext ctx, GravitateDataCommand command) throws Throwable
      {
         return command;
      }

      @Override
      public Object visitEvictFqnCommand(InvocationContext ctx, EvictCommand command) throws Throwable
      {
         EvictCommand clone = commandsFactory.buildEvictFqnCommand(command.getFqn());
         return setDataVersion(clone, command.getFqn());
      }

      private Object setDataVersion(VersionedDataCommand clone, Fqn fqn)
      {
         DataVersion versionToBroadcast = getVersionToBroadcast(workspace, fqn);
         clone.setDataVersion(versionToBroadcast);
         return clone;
      }

      @Override
      public Object visitPutDataMapCommand(InvocationContext ctx, PutDataMapCommand command) throws Throwable
      {
         PutDataMapCommand clone = commandsFactory.buildPutDataMapCommand(null, command.getFqn(), command.getData(), command.isCreateUndoOps(), command.isEraseContents());
         return setDataVersion(clone, clone.getFqn());
      }

      @Override
      public Object visitPutKeyValueCommand(InvocationContext ctx, PutKeyValueCommand command) throws Throwable
      {
         PutKeyValueCommand clone = commandsFactory.buildPutKeyValueCommand(null, command.getFqn(), command.getKey(), command.getValue(), command.isCreateUndoOps(), command.isPutForExternalRead());
         return setDataVersion(clone, command.getFqn());
      }

      @Override
      public Object visitRemoveNodeCommand(InvocationContext ctx, RemoveNodeCommand command) throws Throwable
      {
         RemoveNodeCommand clone = commandsFactory.buildRemoveNodeCommand(command.getGlobalTransaction(), command.getFqn(), command.isEviction(),
               command.isSkipSendingNodeEvents(), command.isCreateUndoOps());
         return setDataVersion(clone, command.getFqn());
      }

      @Override
      public Object visitRemoveKeyCommand(InvocationContext ctx, RemoveKeyCommand command) throws Throwable
      {
         RemoveKeyCommand clone = commandsFactory.buildRemoveKeyCommand(null, command.getFqn(), command.getKey(), command.isCreateUndoOps());
         return setDataVersion(clone, command.getFqn());
      }

      @Override
      public Object visitRemoveDataCommand(InvocationContext ctx, RemoveDataCommand command) throws Throwable
      {
         RemoveDataCommand clone = commandsFactory.buildRemoveDataCommand(command.getGlobalTransaction(), command.getFqn(), command.isCreateUndoops(),
               command.isSendNodeEvent(), command.isEviction());
         return setDataVersion(clone, command.getFqn());
      }

      @Override
      public Object visitCreateNodeCommand(InvocationContext ctx, CreateNodeCommand command) throws Throwable
      {
         return command;
      }

      @Override
      public Object visitMoveCommand(InvocationContext ctx, MoveCommand command) throws Throwable
      {
         return command;
      }

      @Override
      public Object handleDefault(InvocationContext ctx, VisitableCommand command) throws Throwable
      {
         throw new CacheException("Not handling " + command + "!");
      }

      /**
       * Digs out the DataVersion for a given Fqn.  If the versioning is explicit, it is passed as-is.  If implicit, it is
       * cloned and then incremented, and the clone is returned.
       */
      private DataVersion getVersionToBroadcast(TransactionWorkspace w, Fqn f)
      {
         WorkspaceNode n = w.getNode(f);
         if (n == null)
         {
            if (trace) log.trace("Fqn " + f + " not found in workspace; not using a data version.");
            return null;
         }
         if (n.isVersioningImplicit())
         {
            DefaultDataVersion v = (DefaultDataVersion) n.getVersion();
            if (trace)
               log.trace("Fqn " + f + " has implicit versioning.  Broadcasting an incremented version.");

            // potential bug here - need to check if we *need* to increment at all, because of Configuration.isLockParentForChildInsertRemove()
            return v.increment();
         }
         else
         {
            if (trace) log.trace("Fqn " + f + " has explicit versioning.  Broadcasting the version as-is.");
            return n.getVersion();
         }
      }

   }

   protected TransactionWorkspace getTransactionWorkspace(GlobalTransaction gtx) throws CacheException
   {
      OptimisticTransactionEntry transactionEntry = (OptimisticTransactionEntry) txTable.get(gtx);
      if (transactionEntry == null)
      {
         throw new CacheException("unable to map global transaction " + gtx + " to transaction entry");
      }
      // try and get the workspace from the transaction
      return transactionEntry.getTransactionWorkSpace();
   }

}